/***************************************************************************\
* Project Von Neumann - unal.c code file                                    *
* ------------------------------------------------------------------------- *
* Copyright (C) 1995 - 1997 Project Von Neumann Team.  All Rights Reserved. *
* ------------------------------------------------------------------------- *
* Last Revised: 4/8/97                                                      *
* By          : Adrian Tymes                                                *
* Version #   : 0.81                                                        *
* Description : Universe to AL system                                       *
\***************************************************************************/ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "untypes.h"
#include "unal.h"
#include "unutil.h"

/****************************************************************************
 * AL->UN_Object Reference System.
 *
 * ALife nodes can obtain a references to UN_Objects. They can have a
 * reference an arbitary length of time and use it even after the object
 * it refers to is dead. UN has to respond to requests to access a dead
 * object reference meaningfully i.e. it can't reuse a reference.
 *
 * The solution is to never give nodes a pointer to a UN_Object, instead
 * all references are given as AL_ObjectRef's. These are indices into
 * the circular array ref2obj_tb. The array holds the UN_Object
 * pointers. The index for each object is stored in the '.ref' field of
 * the UN_Object and initialised when the object is created.
 *
 * The ref2obj_tb array is circular, headRef and tailRef hold the head
 * and tail indices. These variables are only wrapped when the array is
 * accessed, i.e. they are only ever increased and to access the array
 * they are used 'MOD array_size'. For example, here is the array after
 * the pointers have 'wrapped' once
 *
 *               0   1   2   3   4   5   6   7
 *             ---------------------------------
 *             | x | x | L | x | L | L | x | x | 
 *             ---------------------------------
 *                       ^               ^
 *                    Tail=10         Head=14
 *
 *      x = Dead object
 *      L = Live object
 *
 * When an object is deleted it's entry in the ref2obj_tb is set to
 * point at dummyObj. Currently this is a static object which has
 * suitable for returning from the AL_Get routines. Some of these
 * routines may need to check if the object they are using is dummObj
 * and modify their behaviour accordingly. None of them do just now.
 *
 * If the object being deleted is the tail object in the ref2obj_tb
 * then the tail is moved on to the next live object. E.g. In the
 * diagram above when the tail object dies the Tail will be moved to
 * position 4.
 *
 * The point of this is that whenever AL accesses an object which has
 * a .ref value LESS THAN tailRef then we know that object is dead, so
 * we can just use the dummyObj. Therefore we can respond to accesses
 * to objects created at the start of the game (of course after 2^31
 * objects have been created we will have problems).
 *
 * This is all very well, but what happens if the first object created
 * stays alive for ages? Well, worst case scenario we get this
 *
 *               0   1   2   3   4   5   6   7
 *             ---------------------------------
 *             | L | x | x | x | x | x | L | x | 
 *             ---------------------------------
 *               ^                           ^
 *            Tail=0                      Head=7
 *
 * When the next object is created it would be put in position 7 and
 * head would be moved on. The next object will therefore overwrite
 * the object at position 0! The problem is we can't use any of the
 * dead positions since we may get requests from AL to access those
 * objects. There is currently an assert when a new object is added
 * to check this situation has not occured.
 *
 * NOTE: I have wrapped up all this in some (horrible) macros.
 *
 ****************************************************************************
*/


/***************************************************************************\
* Description: Filescope variables
\***************************************************************************/

/* Tail and Head of the circular ref2obj_tb. Used 'mod REF2OBJTB_SIZE'.
 * Tail is the positon of the oldest alive object.
 * Head is the position to put the _next_ created object.
*/
static AL_ObjectRef tailRef = 0;
static AL_ObjectRef headRef = 0;

/* 'Dummy' object which is accessed when a AL node requests info on
 * a dead object. Used so that some of the AL_Get routines don't need
 * to check if the object is dead (e.g. if we returned NULL instead).
*/
static UN_Object dummyObj = {

        OBJ_DUMMY,  /* object type */
        0, 0,       /* coordinates in universe */
        0, 0,       /* sector we are in */
        0, 0,       /* velocity vector */
        0,          /* facing direction 0..255 */
        1, 1,       /* object weight and height (non-0 incase divide) */
        0,          /* shield strength */
        0,          /* energy/nutricious value/explosive value */
        1,          /* mass of object (non-0 incase divide) */
   0,          /* age */
   NULL, NULL, /* next_sect, prev_set */
        NULL, NULL,  /* next_glob, prev_glob */
        NULL,       /* picture */
        0,          /* reference in ref2obj_tb. Shouldn't be used */
   {
           0,       /* damage */
      0        /* type */
   }
};

/* Pointer to the dummy object
*/
static UN_Object* dummyObjPtr = &dummyObj; 

/* The circular array to hold the object pointer of all alive objects.
 * It's size should be a power of 2 so the compiler _should_ optimize
 * the '% size' to a simple mask operation. We could explicitly do a
 * mask, but there is no real point as all this change no doubt.
*/
#define REF2OBJTB_SIZE 1024

static UN_Object* ref2obj_tb[ REF2OBJTB_SIZE ];

/***************************************************************************\
* Description: Macros to implement reference system
\***************************************************************************/

/* Access the ref2obj_tb. Makes it easy to change.
 * E.g. use a mask instead of mod.
 * IN:
 *   ref = Position in ref2obj_tb to access
*/
#define REF2OBJ_TB(ref) \
        ref2obj_tb[ (ref) % REF2OBJTB_SIZE ]

/* Get a pointer to the object associated with the reference
 * If the reference is < tailRef then we know the object is dead so
 * give a pointer to the dummyObj.
 * IN:
 *   ref = reference of object to return pointer for
 * NOTE:
 *   Don't use parameter with side-effect operators
*/
#define OBJPTR(ref) \
        (((ref) >= tailRef) ? REF2OBJ_TB(ref) : dummyObjPtr)

/* Get the reference field of the given object.
 * IN:
 *   obj = UN_Object pointer to get .ref field of
*/
#define REF(obj) \
        ((obj)->ref)

/****************************************************************************\
* Function Name: UN_AddObjectRef
* Description  : Add the newly created object to the ref2obj_tb
* Arguments    : obj - the object to add
* Returns      : void
\****************************************************************************/
void UN_AddObjectRef(UN_Object* obj)
{
   /* We can't allow probes to obtain references to bullets since
    * there will be just too many. We will probably just allow them
    * to get info on the last bullet that hit them or something.
   */
   /* NOTE: For just just checking it's a bullet, in reality
    * probably want a record of info for each object type (e.g. a
    * UN_ObjTypeInfo struct) and one of the fields would be can this
    * referenced i.e. when probes ask for a list of all objects in
    * the nearby area do we exclude objects of type X.
   */
   if (obj->what == OBJ_BULLET)
   {
      /* Set to -1 so if it was used the would return dummyObj since
       * it will always be less than tailRef.
      */
      obj->ref = -1;
      return;
    }

   /* Set the reference field of the object
   */
        REF(obj) = headRef;

   /* Store the object ptr in the ref2obj_tb
   */
        REF2OBJ_TB(headRef++) = obj;

   /* Check that still space in the ref2obj_tb
   */
        assert((headRef-tailRef) < REF2OBJTB_SIZE );
}

/****************************************************************************\
* Function Name: UN_DelObjectReference
* Description  : Delete the killed object from the ref2obj_tb
* Arguments    : obj - the object to delete
* Returns      : void
\****************************************************************************/
void UN_DeleteObjectRef(UN_Object* obj)
{
   /* See note in AddObjectRef about ignoring certain obect types
   */
   if (obj->what == OBJ_BULLET)
      return;

   /* Set the obj's entry in ref2obj_tb to point to dummy object
   */
        REF2OBJ_TB( REF(obj) ) = dummyObjPtr;

   /* If it's the tail object that just removed then move the
    * the tail on until find a live object or reach the head
    * (in which case there are no alive objects)
   */
        if ( REF(obj) == tailRef )
                while ( (REF2OBJ_TB( ++tailRef ) == dummyObjPtr) &&
                                  (tailRef < headRef))
         ;
}

/****************************************************************************\
* Function Name: AL_Get
* Description  : Get information on the specified object for a AL node.
*                Some routines take two object references, the asker and
*                and the askee.
* Arguments    : ref - reference of object to return info about
*                asker - reference of the object requesting the info
*                askee - reference of the object to return info about
* Returns      : void
\****************************************************************************/

unsigned short AL_GetObjWidth(AL_ObjectRef ref)
{
        return OBJPTR(ref)->w;
}

unsigned short AL_GetObjHeight(AL_ObjectRef ref)
{
        return OBJPTR(ref)->h;
}

UN_ObjType AL_GetObjType(AL_ObjectRef ref)
{       
        return OBJPTR(ref)->what;
}

UN_Velocity AL_GetAbsXVel(AL_ObjectRef ref)
{
        return OBJPTR(ref)->velx;
}

UN_Velocity AL_GetAbsYVel(AL_ObjectRef ref)
{
        return OBJPTR(ref)->vely;
}

UN_Velocity AL_GetAbsVelSqrd(AL_ObjectRef ref)
{
   UN_Object* obj = OBJPTR(ref);
        return ((obj->velx * obj->velx)+(obj->vely * obj->vely));
}

UN_Velocity AL_GetRelXVel(AL_ObjectRef askee, AL_ObjectRef asker)
{
        return ( AL_GetAbsXVel(askee) - AL_GetAbsXVel(asker) );
}

UN_Velocity AL_GetRelYVel(AL_ObjectRef askee, AL_ObjectRef asker)
{
        return ( AL_GetAbsYVel(askee) - AL_GetAbsYVel(asker) );
}

UN_Velocity AL_GetRelVelSqrd(AL_ObjectRef askee, AL_ObjectRef asker)
{
        UN_Velocity x,y;
        x = AL_GetRelXVel(askee, asker);
        y = AL_GetRelYVel(askee, asker);
        return ((x * x)+(y * y));
}

UN_Coord AL_GetAbsXPos(AL_ObjectRef ref)
{
        return OBJPTR(ref)->x;
}

UN_Coord AL_GetAbsYPos(AL_ObjectRef ref)
{
        return OBJPTR(ref)->y;
}

UN_Coord AL_GetRelXPos(AL_ObjectRef askee, AL_ObjectRef asker)
{
        UN_Coord x = AL_GetAbsXPos(askee) - AL_GetAbsXPos(asker);
        if (x > MAX_X/2) return x-=MAX_X;
        else if ((-1)*x > MAX_X/2) return x+=MAX_X;
        else return x;
}

UN_Coord AL_GetRelYPos(AL_ObjectRef askee, AL_ObjectRef asker)
{
        UN_Coord y = AL_GetAbsYPos(askee) - AL_GetAbsYPos(asker);
        if (y > MAX_Y/2) return y-=MAX_Y;
        else if ((-1)*y > MAX_Y/2) return y+=MAX_Y;
        else return y;
}

/* Above if-elses handle wraparound: if two probes are just across the
** sector's "edge" from each other, that shouldn't cause them to think
** that they have to cross the entire sector to meet each other when
** they can just cross the "edge". */

/* Facing means "what direction is an object pointing towards".
** Heading means "what direction is an object travelling in".
** AL_GetRelDirection() returns the direction from one object's
** current position to another's, so no absolute version of that
** function can exist by definition. */

UN_Direction AL_GetAbsFacing(AL_ObjectRef ref)
{
        return OBJPTR(ref)->facedir;
}

UN_Direction AL_GetRelFacing(AL_ObjectRef askee, AL_ObjectRef asker)
{
        UN_Direction x = AL_GetAbsFacing(askee) - AL_GetAbsFacing(asker);
        if (x < 0) return x+NUM_DIRS;
        else return x;
}

UN_Direction AL_GetAbsHeading(AL_ObjectRef ref)
{
        return UN_AntiTanVel( AL_GetAbsXVel(ref), AL_GetAbsYVel(ref) );
}

UN_Direction AL_GetRelHeading(AL_ObjectRef askee, AL_ObjectRef asker)
{
        return UN_AntiTanVel( AL_GetRelXVel(askee,asker),
                         AL_GetRelYVel(askee,asker) );
}

UN_Direction AL_GetRelDirection(AL_ObjectRef askee, AL_ObjectRef asker)
{
   /* Think this is does the the same as the original. Who wrote it? */

   return ( UN_AntiTanVel( AL_GetRelXPos(askee, asker),
                           AL_GetRelYPos(askee, asker) ) );
}

/* Everything after this comment should only be called by the object
** that it refers to.  No other object should be able to directly
** observe this information, although they can figure it out indirectly.
** Functions have been designated ALI_ to highlight this. */

long ALI_GetShield(AL_ObjectRef ref)
{
        return OBJPTR(ref)->shield;
}

long ALI_GetEnergy(AL_ObjectRef ref)
{
        return OBJPTR(ref)->energy;
}

long ALI_GetMass(AL_ObjectRef ref)
{
        return OBJPTR(ref)->mass;
}

unsigned long ALI_GetAge(AL_ObjectRef ref)
{
        return OBJPTR(ref)->age;
}

/* Bullet is not returned since bullets aren't animate.  We'll need
** something to handle this for missiles when we implement them. */
/* SMD: Don't think we will/can allow nodes to have ref's to bullets
** since it could fill the obj2ref_tb too quickly. This has been
** discussed before.
*/

UN_AnimateInfo ALI_GetAnim(AL_ObjectRef ref)
{
        return OBJPTR(ref)->extra.animate;
}

UN_Direction ALI_GetTurnRate(AL_ObjectRef ref)
{
        return OBJPTR(ref)->extra.animate.turn;
}

unsigned long ALI_GetThrustRate(AL_ObjectRef ref)
{
        return OBJPTR(ref)->extra.animate.thrust;
}

UN_Actions ALI_GetActions(AL_ObjectRef ref)
{
        return OBJPTR(ref)->extra.animate.actions;
}
