/***************************************************************************\
* Project Von Neumann - ungame.c code file                                  *
* --------------------------------------------------------------------------*
* Copyright (C) 1995 - 1997 Project Von Neumann Team.  All Rights Reserved. *
* --------------------------------------------------------------------------*
* Last Revised: 4/12/97                                                     *
* By          : Adrian Tymes                                                *
* Version #   : 0.81                                                        *
* Note        : The physics of the game                                     *
\***************************************************************************/ 

/* #define DEBUG for debugging text */
//#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/* for NULL */
#include <stdlib.h>

/* PVN header files */
#include "fe.h"
#include "as.h"
#include "untypes.h"
#include "unutil.h"
#include "unobject.h"
#include "ungame.h"
#include "ai.h"
#include "main.h"

/***************************************************************************\
* Description: Global variables
\***************************************************************************/

UN_Object* Player; /* pointer to the player object */
UN_Object* **Sectors; /* dynamically allocated 2-d array of ptrs to obj 
lists */
UN_Object* ActiveObjects;

/* trig lookup tables */
float UN_Sin[NUM_DIRS];
float UN_Cos[NUM_DIRS];

long ThrustCost[MAX_THRUST];

UN_Coord size_x, size_y;
unsigned long size_sectx, size_secty;
unsigned short shifts_sectx, shifts_secty;

/***************************************************************************\
* Description: Local Functions
\***************************************************************************/

static void UN_UpdateAll();
static void UN_UpdateObject(UN_Object* object);
static void UN_ApplyActions(UN_Object* object);
static void UN_ApplyPhysics(UN_Object* object);
static void UN_ApplyCollisions(UN_Object* object);
static short UN_CheckSystemKeys();
static short UN_CheckForEnd();

/***************************************************************************\
* Function Name: UN_GameLoop
* Description  : The main game loop
* Arguments    : none
* Returns      : The status of the game, as defined in main.c
\***************************************************************************/
short UN_GameLoop()
     {
	short status; 
	FE_Ticks NextUpdate;
	status = 3;
	while (status == 3)
	  { 
	     NextUpdate = FE_Set();
	     UN_UpdateAll(); /* update all objects */
	     FE_Update(Sectors,Player->x,Player->y,
                 shifts_sectx,shifts_secty); /* draw screen */
	     AI_DoQueue(NextUpdate); /* do AI stuff */
	     status = UN_CheckSystemKeys(); /* handle system keys */
	     if (status == 3) status = UN_CheckForEnd();
		 /* see if all probes and/or players are dead */
	  } 
	return status;
     }

/***************************************************************************\
* Function Name: UN_UpdateAll
* Description  : Updates all objects in the universe.
* Arguments    : none
* Returns      : void
\***************************************************************************/
void UN_UpdateAll()
     {
	UN_Object* current;
	
	/* fetch player actions */
	Player->extra.animate.actions = FE_GetKeys(); 
	
	/* fetch probe etc actions here eventually with 
	 * AL functions. Note that these AL functions are not going
	 * to do a lot of processing; Alife will do processing in its own
	 * time segment, semi-independent of universe. */
	
	/* go through active object list, and update all objects in it
	 * Note that active objects list should not be empty here */
	for (current = ActiveObjects->next_glob; /* skip dummy */ 
	     current != NULL; /* as long as we are not at the end */
	     current = current->next_glob) /* next object */
	  {
	     if (current->what == OBJ_PROBE)
		 current->extra.animate.actions = AI_GetKeys(current);
	     UN_UpdateObject(current);
	  }
     }

/***************************************************************************\
* Function Name: UN_UpdateObject
* Description  : Updates everything in an object, physics, collision checking
*                and actions if the object was animate.
* Arguments    : The object to update.
* Returns      : void
\***************************************************************************/
void UN_UpdateObject(UN_Object* object)
     {
	/* If the object is animate, apply its actions */
	if (object->what <= OBJ_MISSILE)
	  {
	     UN_ApplyActions(object);
#ifdef DEBUG
	     printf("Actions:%d velx:%f vely:%f face:%d\n",
		    ANIMATE.actions,
		    object->velx, object->vely, object->facedir);
#endif
	  }
	
	UN_ApplyPhysics(object); /* Do physics */
	UN_ApplyCollisions(object); /* check for collisions */
	if (object->shield <= 0) { /* The object is dead */
		if (object == Player) UN_NewPlayer();
		UN_KillObject(object);
	}
     }

/***************************************************************************\ 
* Function Name: UN_ApplyActions
* Description  : Applies the actions as set in the animate object 
*                (in 'extra.animate.actions'), to that object.
* Arguments    : object - the object we apply the actions to
* Returns      : void
* Notes        : This function assumes the actions (keypresses) are 
*                set in a previous function.
*                The macro ANIMATE stands for object->extra.animate
*                The thrust and brake code currently uses some Allegro
*                fixed point routines (in an inefficient way, too. Just to
*                                      make it work)
\***************************************************************************/
void UN_ApplyActions(UN_Object* object)
     {
	static unsigned long curthrust;  /* temporary holder for thrust */

	if (ANIMATE.actions & ACT_THRUST) /* forward thrust */
	  {
	     curthrust = ANIMATE.thrust; /* fetch thrust from object */
	     if (object->energy > ThrustCost[curthrust]) /*if enough energy*/
	       {
		  object->velx += 
		  ((float)curthrust / (float)object->mass) *
		  UN_Sin[object->facedir];
#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Sin:%f\n",
			 curthrust, object->mass, UN_Sin[object->facedir]);
#endif /* DEBUG */
		  object->vely -= 
		  ((float)curthrust / (float)object->mass) *
		  UN_Cos[object->facedir];
#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Cos:%f\n",
			 curthrust, object->mass, UN_Cos[object->facedir]);
#endif /* DEBUG */
		  
		  /* update energy */
		  object->energy -= ThrustCost[curthrust];
	       }
	     else /* not enough energy for thrust */
	       {
#ifdef DEBUG
		  if (object->what == OBJ_PLAYER)
		      printf("Player can not thrust: not enough energy\n");
#endif /* DEBUG */
		  /* Need to set AS special effect for 
		   * failure of thrust (sound?) */
		  
		  /* hardware interrupt for this object.
		   * Nothing extensive happens here, just preparation for
		   * the Alife handler that's called later. */
		  /* AL_Interrupt(object, THRUST_FAIL); */
	       }
	  } /* end of THRUST code */
   /* we can't thrust and brake at the same time, hence the else if */
   /* NOTE: SMD: Is this true? We don't write the probes programs */
   /* AT: But we do write their bodies.  If they try it, they'll thrust. */     
	else if (ANIMATE.actions & ACT_BRAKE)
	  {
	     curthrust = ANIMATE.thrust;
	     if (object->energy > ThrustCost[curthrust]) /*if enough energy*/
	       {
		  object->velx -= 
		  ((float)curthrust / (float)object->mass) *
		  UN_Sin[object->facedir];

#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Sin:%f\n",
			 curthrust, object->mass, UN_Sin[object->facedir]);
#endif /* DEBUG */

		  object->vely += 
		  ((float)curthrust / (float)object->mass) *
		  UN_Cos[object->facedir];

#ifdef DEBUG
		  printf("curthrust:%lu mass:%lu UN_Cos:%f\n",
			 curthrust, object->mass, UN_Cos[object->facedir]);
#endif /* DEBUG */

		  /* update energy */
		  object->energy -= ThrustCost[curthrust];
	       }
	     else /* not enough energy for brake thrust */
	       {
		  /* Need to set AS special effect for 
		   * failure of brake (sound?) */
		  
		  /* hardware interrupt for this object.
		   * Nothing extensive happens here, just preparation for
		   * the Alife handler that's called later. */
		  /* AL_Interrupt(object, BRAKE_FAIL); */
	       } 
	  } /* end of BRAKE code - end of THRUST/BRAKE code */

	if (ANIMATE.actions & ACT_LEFT) /* turn left */
	  {
	     /* counter clockwise */
	     /* faster version, but less portable:
	      object->facedir = (object->facedir - ANIMATE.turn) & NUM_DIRS-1;
	      */
	     object->facedir = 
	     ((object->facedir > ANIMATE.turn)
	      ? (object->facedir - ANIMATE.turn)
	      : (NUM_DIRS - (ANIMATE.turn - object->facedir)));
	  } /* end of LEFT code */
	else if (ANIMATE.actions & ACT_RIGHT) /* turn right */
	  {
	     /* clockwise */
	     object->facedir = (object->facedir + ANIMATE.turn) & (NUM_DIRS-1);
	  } /* end of RIGHT code */
	
	if (ANIMATE.actions & ACT_FIRE) /* fire currently selected weapon */
	  {
	    if (object->energy > ((UN_Object *)ANIMATE.weapon)->mass) {
		object->energy -= ((UN_Object *)ANIMATE.weapon)->mass;
		UN_FireWeapon(object);
	    } else { /* not enough energy to fire */
	      /* Need to set AS special effect for 
	       * failure of fire (sound?) */

	      /* hardware interrupt for this object.
	       * Nothing extensive happens here, just preparation for
	       * the Alife handler that's called later. */
	      /* AL_Interrupt(object, FIRE_FAIL); */
	    }
	  }

	/* target cycling for player: not yet */

     }
	     
/**************************************************************************\
* Function Name: UN_ApplyPhysics
* Description  : Applies physics to an object. This means an object
*                keeps moving with the same velocity.
* Arguments    : object - what we apply the physics to
* Returns      : void
* Notes        : Collision checking is done in UN_ApplyCollisions
\**************************************************************************/
void UN_ApplyPhysics(UN_Object* object)
     {
	/* Flag, set if we moved out of sector */
	UN_Bool moved_out;
	
	/* Update the age of the object */
	if ((object->what == OBJ_BULLET) ||
	    (object->what == OBJ_MISSILE))
	  {
	     /* bullets and missiles have a maximum lifespan that is
	      * set at their creation. Their age decreases until it
	      * is 0. When the age is 0, their shield is set to 0,
	      * meaning the object will now die. */
	     (object->age)--;
	     if (object->age <= 0)
	       object->shield = 0;
	  }
	else
	  {
	     /* other objects have increasing ages */
	     (object->age)++; 
	  }
	
	/* update the x and y of the object */
	
	object->x += object->velx;
	object->y += object->vely;
	
	/* it possible we now moved out of the playing field 
	 * For ease of coding, UN_Coord is *signed*. 
	 * This means the max field is 2 to the power 31 large, assuming
	 * 32 bit longs. This is about 2,147,000,000 maximum.  
	 * Large enough. */
	
	/* NOTE: we are now using wraparound walls */
	
	  /* if we moved out on the left, move in from the right */
	if (object->x < 0)
	  object->x = size_x + object->x;

	  /* if we moved out on the top, move in from the bottom */
	if (object->y < 0)
	  object->y = size_y + object->y;
	
	  /* if we moved out on the right, move in from the left */
	if (object->x >= size_x)
	  object->x = object->x - size_x;
	
	  /* if we moved out on the bottom, move in from the top */
	if (object->y >= size_y)
	  object->y = object->y - size_y; 
	
	/* now check if we moved out of our sector */
	moved_out = FALSE;
	if (object->x / size_sectx != object->sectx)
	  {
	     object->sectx = object->x / size_sectx;
	     moved_out = TRUE;
	  }
	
	if (object->y / size_secty != object->secty)
	  {
	     object->secty = object->y / size_secty;
	     moved_out = TRUE;
	  }
	
	/* if we indeed moved out of our sector, update object lists */
	if (moved_out)
	  UN_ShiftObject(object);               
     }

/**************************************************************************\
* Function Name: UN_ApplyCollisions
* Description  : Checks for any collisions with object and resolves them
* Arguments    : object - what we check collisions against
* Returns      : void
\**************************************************************************/
void UN_ApplyCollisions(UN_Object* object)
     {
	 /* current object to check for collision with */
	 UN_Object* current;
	 /* current sector we're checking in */
	 UN_Sector i, j;

	 short k, l;
	 short numx, numy;
	 UN_Coord corx, cory;
	 UN_Velocity tempx, tempy;

	 corx = cory = 0;

	 /* top left sector to check */
	 i = object->sectx;
	 j = object->secty;

	 /* set number of sectors to look through */
	 numx = ((object->x + object->w) / size_sectx) - i;

	 numy = ((object->y + object->h) / size_secty) - j;

	 for (k=0; k <= numx; k++)
	  {
	   for (l=0; l <= numy; l++)
	    {
	     for (current = Sectors[i][j]->next_sect; /* skip dummy */
		    current != NULL; /* check each object in this sector*/
		    current = current->next_sect)
		{
		 /* can't collide with self and bullet explosion can only
		    affect one object */
		 if ((current == object) || (current->shield <= 0)) continue;

		 /* check for collision */
		 if ((current->x + corx) > (object->x - current->w) &&
		     (current->y + cory) > (object->y - current->h) &&
		     (current->x + corx) < (object->x + object->w) &&
		     (current->y + cory) < (object->y + object->h))
		  {
		   /* collision has occured, now handle it */
		   /* BH: Maybe this section should be a seperate function*/
		   /* first we check for explosives */
		   if ((object->what == OBJ_BULLET) ||
			 (object->what == OBJ_MISSILE) ||
			 (current->what == OBJ_BULLET) ||
			 (current->what == OBJ_MISSILE))
		    {
		     if ((object->what == OBJ_BULLET) ||
			   (object->what == OBJ_MISSILE))
			{
			 current->shield -= object->extra.bullet.damage;
			 object->shield = 0;
			 /* need calls to AS for explosions */
			}
		     if ((current->what == OBJ_BULLET) ||
			   (current->what == OBJ_MISSILE))
			{
			 object->shield -= object->extra.bullet.damage;
			 current->shield = 0;
			 /* need calls to AS for explosions */
			}
		    }
		   /* need to add checks for everything else */
		   else
		    { /* objects bounce off of each other */
		     tempx = object->velx * object->mass;
		     tempy = object->vely * object->mass;

		     object->velx = (current->velx * current->mass)
				    / object->mass;
		     object->vely = (current->vely * current->mass)
				    / object->mass;

		     current->velx = tempx / current->mass;
		     current->vely = tempy / current->mass;
		    }
		  }
		}

	     /* reset the y correcting variable */ 
	     cory = 0;

	     /* get the sector y, correct for wraparound if necessary */
	     if ((++j) >= (size_y/size_secty))
		{
		 j=0;
		 corx = size_y;
		}
	    }

	   /* reset the correcting variables */
	   corx = cory = 0;
	   /* reset the sector y */
	   j = object->secty;

	   /* get the sector x, correct for wraparound if necassary */
	   if ((++i) >= (size_x/size_sectx))
	    {
	     i=0;
	     cory = size_x;
	    }
	  }
     }

/****************************************************************************\
* Function Name: UN_CheckSystemKeys
* Description  : Checks to see if any system keys are being pressed, and if
*                so, takes the appropriate action.
* Arguments    : None
* Returns      : The status of the system:
*                3 = keep playing, 2 = pause, 1 = quit to main menu, 0 = quit
\****************************************************************************/
short UN_CheckSystemKeys() {
    UN_Actions in = FE_GetKeys();
    if ((in & SYS_TOGGLE_MUSIC) || (in & SYS_TOGGLE_SFX)) FE_ToggleSound();
    if (in & SYS_QUIT) return 0;
    else if (in & SYS_END_GAME) return 1;
    else if (in & SYS_PAUSE) return 2;
    else return 3;
}

/****************************************************************************\
* Function Name: UN_CheckForEnd
* Description  : Checks to see if any of the following conditions are met:
*                Player is dead - game over
*                All probes are dead - advance to next wave or end game
*                Plays the appropriate movie if these conditions are met.
* Arguments    : None
* Returns      : 1 (quit to main menu) if the game is over, otherwise 3 (keep
*                playing)
\****************************************************************************/
short UN_CheckForEnd() {
    UN_Object *current;
    if (Player->what == OBJ_D_PLAYER) { /* Player is dead */
	AS_ShowMovie(GAME_OVER);
	return 1;
    } else for (current = ActiveObjects->next_glob; /* skip dummy */ 
	    current != NULL; /* as long as we are not at the end */
	    current = current->next_glob) /* next object */
	if (current->what == OBJ_PROBE) return 3;
    /* if we get here, there are no probes remaining */
    curmovie = curmovie + 1;
    if (curmovie == 2) { /* Or however many waves we want to have */
	AS_ShowMovie(PLAYER_WON);
	return 1;
    } else { /* Start next wave */
	AS_ShowMovie(WAVE_BASE + curmovie);
	UN_Restock(curmovie);
	return 3;
    }
}
	     

