/***************************************************************************\
* Project Von Neumann - unobject.c code file                                *
* --------------------------------------------------------------------------*
* Copyright (C) 1995 - 1997 Project Von Neumann Team.  All Rights Reserved. *                   
* --------------------------------------------------------------------------*
* Last Revised: 4/8/97                                                      *
* By          : Adrian Tymes                                                *
* Version #   : 0.81                                                        *
* Description : Universe initialisation and object management functions     *
\***************************************************************************/

#include <stdio.h>
#include <math.h>

/* PVN header files */
#include "untypes.h"
#include "unobject.h"
#include "unutil.h"
#include "ungame.h"
#include "unal.h"
#include "main.h"
#include "fe.h"
#include "as.h"
#include "ai.h"

/***************************************************************************\
* Description: Filescope Variables
\***************************************************************************/

static UN_Object bullet_prototype;

/***************************************************************************\
* Description: Local Functions
\***************************************************************************/

static void UN_InitSectors(UN_Sector x_sectors, UN_Sector y_sectors);

static void UN_InitObj(UN_Object* object);
static void UN_InitObjCoord(UN_Object* object, UN_Coord x, UN_Coord y,
                     unsigned int w, unsigned int h,
                     UN_Direction facedir);
static void UN_InitObjVelocity(UN_Object* object,
           UN_Velocity velx, UN_Velocity vely);
static void UN_InitObjType(UN_Object* object, UN_ObjType type);
static void UN_InitObjStats(UN_Object* object,
           long shield, long energy, long mass);
static void UN_InitObjPic(UN_Object* object);
static void UN_InitObjAnimate(UN_Object* object, unsigned char turn, 
                       unsigned long thrust, UN_Actions actions,
                       UN_Object *weapon);

static void UN_InitTrig();
static void UN_InitThrust();

/****************************************************************************\
* Function Name: UN_Init
* Description  : Initializes structures and variables needed by universe
*                which do not change when a new game is begun.
* Arguments    : none
* Returns      : void
\****************************************************************************/
void UN_Init() {
    /* initialize UN_Sin & UN_Cos tables */
    UN_InitTrig();
    
    /* initialize thrust cost lookup table */
    UN_InitThrust();
    
    /* initialize default bullet prototype */
    bullet_prototype.mass = 10;
    bullet_prototype.w = 32;
    bullet_prototype.h = 32;
    bullet_prototype.shield = 10;
    bullet_prototype.energy = 10;
    bullet_prototype.what = OBJ_BULLET;
    bullet_prototype.extra.bullet.type = 1;
    bullet_prototype.extra.bullet.damage = 10;

    ActiveObjects = NULL;
}

/****************************************************************************\
* Function Name: UN_InitGame
* Description  : Initializes structures and variables needed by universe
*                which are reset when a new game is begun.
* Arguments    : none
* Returns      : void
\****************************************************************************/
void UN_InitGame()
     {
        UN_Object* other1;
        UN_Object* other2;
        UN_Object* other3;
        UN_Object bullet;        

        /* curmovie is the wave number; at the start of a game, this is 0 */
        curmovie = 0;

        /* temporary size, these should be arguments of something */
        size_x      = 65536 * 64;
        size_y      = 65536 * 64;
        size_sectx  = 65536;
        size_secty  = 65536;
        shifts_sectx = 16;
		shifts_secty = 16;

        /* initialize sectors */
        UN_InitSectors(size_x / size_sectx, size_y / size_secty);
        
        /* initialize active objects list */
        while (ActiveObjects != NULL) {
            /* Erase previous game's active objects */
            other1 = ActiveObjects;
            ActiveObjects = ActiveObjects->next_glob;
            UN_DeleteObject(other1);
        }
        ActiveObjects = UN_NewObject();
        ActiveObjects->next_glob = NULL;
        ActiveObjects->prev_glob = NULL;
        ActiveObjects->what = OBJ_DUMMY;
        
        /* make the player */
        Player = UN_NewObject();
        
        /* -- Test initialisation follows -- */
        
        /* initialize the player */
        UN_InitObj(Player);
        UN_InitObjType(Player, OBJ_PLAYER);
        UN_InitObjCoord(Player, 1000, 1000, 32 * 64, 32 * 64, 0);
        UN_InitObjVelocity(Player, 0, 0);
        UN_InitObjStats(Player, 100, 100000, 10);
        UN_InitObjPic(Player);
        UN_InitObjAnimate(Player, 1, 10, 0, (UN_Object *) &bullet_prototype);
        AI_InitProbeGenome(Player, (AI_Genome *) NULL, (AI_Genome *) NULL);        
                
        /* make a few other objects */
        other1 = UN_NewObject();
        other2 = UN_NewObject();
        other3 = UN_NewObject();

        UN_InitObj(other1);
        UN_InitObjType(other1, OBJ_PROBE);
        UN_InitObjCoord(other1, 2000, 2000, 32 * 64, 32 * 64, 0);
        UN_InitObjVelocity(other1, 0, 0);
        UN_InitObjStats(other1, 100, 1000, 10);
        UN_InitObjPic(other1);
        UN_InitObjAnimate(other1, 1, 1, 0, (UN_Object *) &bullet_prototype); 
        AI_InitProbeGenome(other1, (AI_Genome *) NULL, (AI_Genome *) NULL);        

        UN_InitObj(other2);
        UN_InitObjType(other2, OBJ_PROBE);
        UN_InitObjCoord(other2, 4000, 2000, 32 * 64, 32 * 64, 30);
        UN_InitObjVelocity(other2, 0, 0);
        UN_InitObjStats(other2, 100, 1000, 10);
        UN_InitObjPic(other2);
        UN_InitObjAnimate(other2, 1, 1, 0, (UN_Object *) &bullet_prototype);
        AI_InitProbeGenome(other2, (AI_Genome *) NULL, (AI_Genome *) NULL);        
        
        UN_InitObj(other3);
        UN_InitObjType(other3, OBJ_ASTEROID);
        UN_InitObjCoord(other3, 2000, 4000, 32 * 64, 32 * 64, 200);
        UN_InitObjVelocity(other3, 0, 0);
        UN_InitObjStats(other3, 100, 1000, 10);
        UN_InitObjPic(other3);
        AI_InitOtherGenome(other3);
        
        /* activate everything */
        UN_ActivateObject(Player);
        UN_ActivateObject(other1);
        UN_ActivateObject(other2);
        UN_ActivateObject(other3);
        
        /* put everything in their own sectors */
        UN_ToSector(Player);
        UN_ToSector(other1);
        UN_ToSector(other2);
        UN_ToSector(other3);
     }   

/*************************************************************************\
* Function Name: UN_Restock
* Description  : Restock the probe population at the beginning of a new wave
* Arguments    : The current wave's number
* Returns      : void
\*************************************************************************/
void UN_Restock(short wave_number) {
        UN_Object* new_probe;
        short i;
        for (i = 0; i < wave_number; i++) {
                new_probe = UN_NewObject();
                UN_InitObj(new_probe);
                UN_InitObjType(new_probe, OBJ_PROBE);
                UN_InitObjCoord(new_probe, i*1000, i*1000, 32 * 64, 32 * 64, 0);
                UN_InitObjVelocity(new_probe, 0, 0);
                UN_InitObjStats(new_probe, 100, 1000, 1);
                UN_InitObjPic(new_probe);
                UN_InitObjAnimate(new_probe, 1, 1, 0,
                        (UN_Object *) &bullet_prototype);
                AI_InitProbeGenome(new_probe, (AI_Genome *) NULL,
                        (AI_Genome *) NULL);
                UN_ActivateObject(new_probe);
                UN_ToSector(new_probe);
        }
}

/*************************************************************************\
* Function Name: UN_FireWeapon
* Description  : Creates a bullet (or missile) from the designated object's
*                current weapon
* Arguments    : The firing object
* Returns      : void
\*************************************************************************/
void UN_FireWeapon(UN_Object *firer) {
        UN_Object *new_bullet = UN_NewObject();
        UN_InitObj(new_bullet);
        if (firer->what == OBJ_MISSILE) {
                UN_InitObjType(new_bullet,
                        ((UN_Object *)firer->extra.missile.animate.weapon)->what);
                UN_InitObjCoord(new_bullet, firer->x + (firer->velx * 2),
                        firer->y + (firer->vely * 2),
                        ((UN_Object *)firer->extra.missile.animate.weapon)->w,
                        ((UN_Object *)firer->extra.missile.animate.weapon)->h,
                        firer->facedir);
                UN_InitObjVelocity(new_bullet, firer->velx, firer->vely);
                UN_InitObjStats(new_bullet,
                     ((UN_Object *)firer->extra.missile.animate.weapon)->shield,
                     ((UN_Object *)firer->extra.missile.animate.weapon)->energy,
                     ((UN_Object *)firer->extra.missile.animate.weapon)->mass);
                UN_InitObjPic(new_bullet);
                if (new_bullet->what == OBJ_MISSILE) {
                        UN_InitObjAnimate(new_bullet,
                                ((UN_Object *)firer->extra.missile.animate.weapon)->extra.missile.animate.turn,
                                ((UN_Object *)firer->extra.missile.animate.weapon)->extra.missile.animate.thrust,
                                0,
                                (UN_Object *)((UN_Object *)firer->extra.missile.animate.weapon)->extra.missile.animate.weapon);
                        new_bullet->extra.missile.bullet =
                                ((UN_Object *)firer->extra.missile.animate.weapon)->extra.bullet;
                } else new_bullet->extra.bullet =
                        ((UN_Object *)firer->extra.missile.animate.weapon)->extra.bullet;
                AI_InitProbeGenome(new_bullet, (AI_Genome *) NULL,
                        (AI_Genome *) NULL);
        } else {
                UN_InitObjType(new_bullet,
                        ((UN_Object *)firer->extra.animate.weapon)->what);
                UN_InitObjCoord(new_bullet, firer->x + (firer->velx * 2),
                        firer->y + (firer->vely * 2),
                        ((UN_Object *)firer->extra.animate.weapon)->w,
                        ((UN_Object *)firer->extra.animate.weapon)->h,
                        firer->facedir);
                UN_InitObjVelocity(new_bullet, firer->velx, firer->vely);
                UN_InitObjStats(new_bullet,
                        ((UN_Object *)firer->extra.animate.weapon)->shield,
                        ((UN_Object *)firer->extra.animate.weapon)->energy,
                        ((UN_Object *)firer->extra.animate.weapon)->mass);
                UN_InitObjPic(new_bullet);
                if (new_bullet->what == OBJ_MISSILE) {
                        UN_InitObjAnimate(new_bullet,
                                ((UN_Object *)firer->extra.animate.weapon)->extra.missile.animate.turn,
                                ((UN_Object *)firer->extra.animate.weapon)->extra.missile.animate.thrust,
                                0,
                                (UN_Object *)((UN_Object *)firer->extra.animate.weapon)->extra.missile.animate.weapon);
                        new_bullet->extra.missile.bullet =
                                ((UN_Object *)firer->extra.animate.weapon)->extra.bullet;
                } else new_bullet->extra.bullet =
                        ((UN_Object *)firer->extra.animate.weapon)->extra.bullet;
                AI_InitProbeGenome(new_bullet, (AI_Genome *) NULL, (AI_Genome *) NULL);
        }
        UN_ActivateObject(new_bullet);
        UN_ToSector(new_bullet);
}

/*************************************************************************\
* Function Name: UN_InitSectors
* Description  : Create a 2 dimensional array of sectors (pointers to objects)
* Arguments    : x_sectors, y_sectors - width and height of the 2-d array  
* Returns      : void
* Notes        : the sector is filled with dummy objects.
\*************************************************************************/
static void UN_InitSectors(UN_Sector x_sectors, UN_Sector y_sectors)
     {
        UN_Sector i, j;
        UN_Object* dummy;
        
        Sectors = (UN_Object* **) FE_Malloc(sizeof(UN_Object**) * x_sectors);
        
        for (i = 0; i < x_sectors; i++)
          Sectors[i] = (UN_Object* *) FE_Malloc(sizeof(UN_Object*) * y_sectors);

        /* NOTE: we might need to check if FE_Malloc returns NULL */
        
        /* the sector is filled with dummy objects */
        for (i = 0; i < x_sectors; i++)
          {       
            for (j = 0; j < y_sectors; j++)
               {
                 /* create new dummy object */
                  dummy = UN_NewObject();
                  /* initialize it */
                  dummy->sectx = i;
                  dummy->secty = j;
                  dummy->prev_sect = NULL;
                  dummy->next_sect = NULL;
                  /* let the sectors list point at it */
                  Sectors[i][j] = dummy;
               }
          }
        
     }

/****************************************************************************\
* Function Name: UN_InitObj
* Description  : Initializes an object
* Arguments    : object - which object
* Returns      : void
\****************************************************************************/
static void UN_InitObj(UN_Object* object)
     {
        object->age = 0;
        object->next_sect = NULL;
        object->prev_sect = NULL;
        object->next_glob = NULL;
        object->prev_glob = NULL;
     }

/****************************************************************************\
* Function Name: UN_InitObjCoord
* Description  : Initializes the location of an object, size, 
*                and the way it faces
* Arguments    : object - which object
*                x, y - coordinates
*                w, h - size
*                facedir - way it faces
* Returns      : void
\****************************************************************************/
static void UN_InitObjCoord(UN_Object* object, UN_Coord x, UN_Coord y,
                  unsigned int w, unsigned int h,
                  UN_Direction facedir)
     {
        object->x = x;
        object->y = y;
        
        /* check to make sure 0 <= object->x <= size_x */
        if (object->x < 0) object->x = size_x + object->x;
        else if (object->x >= size_x) object->x = object->x - size_x;
        
        /* check to make sure 0 <= object->y <= size_y */  
        if (object->y < 0) object->y = size_y + object->y;
        else if (object->y >= size_y) object->y = object->y - size_y; 
        
        object->sectx = object->x / size_sectx;
        object->secty = object->y / size_secty;
        object->w = w;
        object->h = h;
        object->facedir = facedir;
     }

/****************************************************************************\
* Function Name: UN_InitObjVelocity
* Description  : Initializes the velocity of an object
* Arguments    : object - which object
*                velx, vely - velocity vector
* Returns      : void
\****************************************************************************/
static void UN_InitObjVelocity(UN_Object* object,
        UN_Velocity velx, UN_Velocity vely)
     {
        object->velx = velx;
        object->vely = vely;
     }

/****************************************************************************\
* Function Name: UN_InitObjType
* Description  : Initializes the type of an object
* Arguments    : object - which object
*                type   - what type
* Returns      : void
\****************************************************************************/
static void UN_InitObjType(UN_Object* object, UN_ObjType type)
     {
        object->what = type;

   /* Add reference to object in the UN->AL obj sys. Must be done
    * after the type has been set since certain objects types are
    * ignored.
   */
   UN_AddObjectRef(object);
     }

/****************************************************************************\
* Function Name: UN_InitObjStats
* Description  : Initializes the starting stats of an object.
* Arguments    : object - which object
*                shield - amount of shield
*                energy - amount of energy
*                mass   - mass of this object
* Returns      : void
\****************************************************************************/
static void UN_InitObjStats(UN_Object* object,
        long shield, long energy, long mass)
     {
        object->shield = shield;
        object->energy = energy;
        object->mass = mass;
     }

/****************************************************************************\
* Function Name: UN_InitObjPic
* Description  : Initializes the picture of an object - a shell for AS 
*                functions
* Arguments    : only the object, which should already contain the needed
*                information
* Returns      : void
\****************************************************************************/
static void UN_InitObjPic(UN_Object* object)
     {
        AS_InitializePic(object);
        AS_GeneratePic(object);
     }

/**************************************************************************\
* Function Name: UN_InitObjAnimate
* Description  : Initializes the animate part of an object
* Arguments    : turn - how fast this object turns
*                thrust - how much thrust power this object has
*                actions - initial action bitstring of this object
*                weapon - default weapon to fire
\**************************************************************************/
static void UN_InitObjAnimate(UN_Object* object, unsigned char turn, 
                       unsigned long thrust, 
                       UN_Actions actions,
                       UN_Object *weapon)
     {
        ANIMATE.turn = turn; 
        ANIMATE.thrust = thrust;
        ANIMATE.actions = actions;
        ANIMATE.weapon = weapon;
     }

/****************************************************************************\
* Function Name: UN_InitThrust
* Description  : Initializes the ThrustCost lookup table
* Arguments    : none
* Returns      : void
\****************************************************************************/
static void UN_InitThrust()
     {
        long i;
        
        for (i = 0; i < MAX_THRUST; i++)
          ThrustCost[i] = i;
     }

/***************************************************************************\
* Function Name: UN_InitTrig
* Description  : Initializes UN_Sin and UN_Cos arrays.
* Arguments    : none
* Returns      : void
\***************************************************************************/
static void UN_InitTrig()
     {
        short i;
        double j;

        for (i = 0; i < NUM_DIRS; i++)
          {
             j = (i * (double)TWOPI) / (double)NUM_DIRS;
             UN_Sin[i] = sin(j);
             UN_Cos[i] = cos(j);
          }
     }
