/***************************************************************************\
* Project Von Neumann - untypes.h header file                               *
* --------------------------------------------------------------------------*
* Copyright (C) 1995 - 1997 Project Von Neumann Team.  All Rights Reserved. *
* --------------------------------------------------------------------------*
* Last Revised: 1/1/97                                                      *
* By          : Adrian Tymes                                                *
* Version #   : 0.80                                                        *
* Description : General stuff used by universe fuctions.                    *
\***************************************************************************/

#ifndef UNTYPES_H
#define UNTYPES_H

/***************************************************************************\
* Description: Universe #defines
\***************************************************************************/

/* TRUE and FALSE need to be defined if they are not defined already */
#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif 

/* ABS is not a standard function so define it here. May need to
 * do check for it already existing as a function or macro
*/
#define ABS(a) ((a) < 0 ? (0-(a)) : (a))

/* The action keys */
#define ACT_THRUST       0x00000001
#define ACT_BRAKE        0x00000002
#define ACT_LEFT         0x00000004
#define ACT_RIGHT        0x00000008
#define ACT_NEXT_TARGET  0x00000010
#define ACT_PREV_TARGET  0x00000020
#define ACT_FIRE         0x00000040
#define ACT_EAT          0x00000080
#define ACT_NEXT_WEAPON  0x00000100
#define ACT_PREV_WEAPON  0x00000200
#define ACT_BREED        0x00000400

/* The system keys.  Only the player can use these.  If the probes try to
 * use them, nothing will happen. */
#define SYS_QUIT         0x80000000
#define SYS_END_GAME     0x40000000
#define SYS_PAUSE        0x20000000
#define SYS_TOGGLE_MUSIC 0x10000000
#define SYS_TOGGLE_SFX   0x08000000

/* For ease of programming */
#define ANIMATE (object->extra.animate)

#define NUM_DIRS 256 /* amount of facing directions */

#define TWOPI 6.2831853

/* Max thrust. Currently it ranges from 0 to 100 
 * Note that we may change this to a parameter eventually. */
#define MAX_THRUST 100

/* Maximum sizes of X, Y values
 * This is if unsigned long (a UN_coord) is 32 bits long */
#define MAX_X ((UN_Coord) 4294967295)
#define MAX_Y ((UN_Coord) 4294967295)

/**************************************************************************\
* Description: Various typedefs
\**************************************************************************/

typedef unsigned long UN_Actions;   /* 32 bit number which stores actions */
typedef signed long UN_Coord;       /* universe coordinates max about 2^31 */
typedef float UN_Velocity;          /* object velocity */
typedef signed short UN_Direction; /* object direction */
typedef unsigned char UN_ObjType;   /* object type */
typedef unsigned short UN_Sector;    /* sector of playing field */
typedef unsigned char UN_Bool;      /* bool */

/* An AL_ObjectRef is the type an AL node uses to hold a reference
 * to a UN_Object. It is signed for safety because sometimes do
 * arithmetic on it. This is used in the UN->AL system in UNAL.c
*/
typedef signed int AL_ObjectRef;

/***************************************************************************\
* Description: Enumeration of the object types.
\***************************************************************************/
enum  {
   OBJ_PLAYER = 0,
   OBJ_PROBE,
   OBJ_SUPPLY,
   OBJ_MISSILE, /* note that everything <= MISSILE is animate */
   OBJ_BULLET,
   OBJ_ASTEROID,
   OBJ_UF_EGG,
   OBJ_F_EGG,
   OBJ_D_EGG,
   OBJ_D_PLAYER,
   OBJ_D_PROBE,
   OBJ_D_SUPPLY,
   OBJ_UNMOVABLE,
   OBJ_STICKY,
   OBJ_DUMMY, 
   OBJ_L_OTYPE
};

/**************************************************************************\
* Description: Structures used by universe
\**************************************************************************/

/* Extra information for animate objects */
typedef struct UN_AnimateInfo_s
     {
        void *weapon; /* a prototype of the current selected weapon */
        UN_Direction turn;
        unsigned long thrust;
        UN_Actions actions;
        void *genome; /* used by AI */
     } UN_AnimateInfo;

/* Extra information for bullet objects */
typedef struct UN_BulletInfo_s
     {
   long damage;
   long type;
     } UN_BulletInfo;

typedef struct UN_MissileInfo_s
     {
   UN_BulletInfo bullet;
   UN_AnimateInfo animate;
     } UN_MissileInfo;

/* A universe object */
typedef struct UN_Object_s
     {
        UN_ObjType what; /* object type */
        UN_Coord x, y;  /* coordinates in universe */
        UN_Sector sectx, secty; /* sector we are in */
        UN_Velocity velx, vely; /* velocity vector */
        UN_Direction facedir; /* facing direction 0..255 */
        unsigned short w, h; /* object weight and height */
        long shield; /* shield strength */
        long energy; /* energy/nutricious value/explosive value */
        long mass;   /* mass of object */
        unsigned long age; /* age */

        /* this object is in a sector list */
        struct UN_Object_s* next_sect;
        struct UN_Object_s* prev_sect;
        /* this object is in the global active object list */
        struct UN_Object_s* next_glob;
        struct UN_Object_s* prev_glob;
        
        void* pic; /* pointer to the current picture */

        AL_ObjectRef ref; /* reference in ref2obj_tb. See unal.c */
        
        /* extra information, specific to object type */
        union
          {
             UN_AnimateInfo animate;
             UN_BulletInfo bullet;
             UN_MissileInfo missile;
          } extra;
        
   } UN_Object;

#endif
