/***************************************************************************\
* Project Von Neumann - unutil.c code file                                  *
* --------------------------------------------------------------------------*
* Copyright (C) 1995 - 1997 Project Von Neumann Team.  All Rights Reserved. *
* --------------------------------------------------------------------------*
* Last Revised: 4/8/97                                                      *
* By          : Adrian Tymes                                                *
* Version #   : 0.81                                                        *
* Description : Utility functions used by universe                          *
\***************************************************************************/

/* if DEBUG is defined, some functions will emit text. */
/* #define DEBUG */

/***************************************************************************\
* Description: header files
\***************************************************************************/

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/* Standard header files */
#include <stdlib.h> 

/* PVN header files */
#include "unutil.h"
#include "untypes.h"
#include "ungame.h"
#include "fe.h"
#include "as.h"

/***************************************************************************\
* Description: Filescope variables
\***************************************************************************/

static UN_Object* SpareStack = NULL; /* spare object stack */

/***************************************************************************\
* Utility functions
\***************************************************************************/
  
/***************************************************************************\ 
* Function Name: UN_NewObject
* Description  : Obtain a new UN_Object
* Arguments    : none  
* Returns      : pointer to newly allocated UN_Object
\***************************************************************************/
UN_Object* UN_NewObject()
     {
        UN_Object* object;
        
        if (SpareStack != NULL)
          {
             /* SpareStack is not empty */
             object = SpareStack;
             SpareStack = SpareStack->next_sect;
             return (UN_Object*) object;
          }
        else
          {
             /* SpareStack is empty, allocate a new object */
             object = (UN_Object*) FE_Malloc(sizeof(UN_Object));
             /* NOTE: we want to check for NULL being returned.
              * (as then no allocation took place) */
             /* Also, AS requires object->pic to be set to NULL here. */
             if (object != NULL) ((UN_Object*)object)->pic = NULL;
             return (UN_Object*) object;
          }
     }

/***************************************************************************\
* Function Name: UN_DeleteObject
* Description  : Delete an object
* Arguments    : object - the object to delete
* Returns      : void
* Notes        : this assumes all object lists start with a dummy object.
\***************************************************************************/
void UN_DeleteObject(UN_Object* object)
     {
   /* Remove the object from the UN->AL obj ref sys
   */
   UN_DeleteObjectRef(object);

        /* Remove object from the sector list it was in */
        
        /* First remove the object from the old list */
        if (object->next_sect != NULL) /* if we are not last object in list */
          {
             /* connect next_sect object to prev_sect object */
             object->next_sect->prev_sect = object->prev_sect; 
          }
        /* connect prev_sect object to next_sect object */
        object->prev_sect->next_sect = object->next_sect;
        
        /* remove the object from the active object list it was in */
        if (object->next_glob != NULL) /* if we are not last object in list */
          {
             /* connect next_glob object to prev_glob object */
             object->next_glob->prev_glob = object->prev_glob;
          }
        /* connect prev_glob object to next_glob object */
        object->prev_glob->next_glob = object->next_glob;
         
        /* Proceed to put object on the top of the spare stack */
        object->next_sect = SpareStack;
        SpareStack = object;
     }

/***************************************************************************\
* Function Name: UN_ToSector(UN_Object* object)
* Description  : Links an object into a sector list.
* Arguments    : object - the object
* Returns      : void
\***************************************************************************/
void UN_ToSector(UN_Object* object)
     {
        /* we do not have to remove the object from anything */
        
        /* Insert the object at the top of the new sector list,
         * under the dummy (which is always the first object) */
        
        /* connect this object to start of rest of list */
        object->next_sect = Sectors[object->sectx][object->secty]->next_sect;
        if (object->next_sect != NULL) /* if we are not last object in list */
          {
             /* connect next_object to this object */
             object->next_sect->prev_sect = object;
          }
        /* connect this object to start of list (the dummy) */
        object->prev_sect = Sectors[object->sectx][object->secty];
        /* connect dummy to this object */
        Sectors[object->sectx][object->secty]->next_sect = object;
        
#ifdef DEBUG
        printf("UN_ToSector: sectx:%lu secty:%lu\n", 
               object->sectx, object->secty);
#endif /* DEBUG */
     }
/***************************************************************************\
* Function Name: UN_ShiftObject
* Description  : Removes the object from the sector object list it is in, and 
*                then puts it in the object list of a new sector.
* Arguments    : object - the object to move
* Returns      : void
\***************************************************************************/
void UN_ShiftObject(UN_Object* object)
     {
        /* First remove the object from the old list */
        
        if (object->next_sect != NULL) /* if we are not last object in list */
          {
             /* connect next_sect object to prev_sect object */
             object->next_sect->prev_sect = object->prev_sect;  
          }
        /* connect prev_sect object to next_sect object */
        object->prev_sect->next_sect = object->next_sect; 
        
        /* Then insert the object at the top of the new sector list,
         * under the dummy (which is always the first object) */
        
        /* connect this object to start of rest of list */
        object->next_sect = Sectors[object->sectx][object->secty]->next_sect;
        if (object->next_sect != NULL) /* if we are not last object in list */
          {
             /* connect next_object to this object */
             object->next_sect->prev_sect = object;
          }
        /* connect this object to start of list (the dummy) */
        object->prev_sect = Sectors[object->sectx][object->secty];
        /* connect dummy to this object */
        Sectors[object->sectx][object->secty]->next_sect = object;
        
#ifdef DEBUG
        printf("UN_ShiftObject: sectx:%lu secty:%lu\n", 
               object->sectx, object->secty);
#endif /* DEBUG */
     }

/***************************************************************************\
* Function Name: UN_ActivateObject
* Description  : Adds an object to top of the ActiveObjects list 
*                (the active objects list) 
* Arguments    : the object to activate
* Returns      : void
* Notes        : Assumes ActiveObjects list has a dummy object on top.
\***************************************************************************/
void UN_ActivateObject(UN_Object* object)
     {
        /* connect next_glob of this object to previous top of the list */
        object->next_glob = ActiveObjects->next_glob;
        if (object->next_glob != NULL) /* if we are not last in list */
          {
             /* connect next object to this object */
             object->next_glob->prev_glob = object;
          }
        
        /* connect this object to the dummy object */
        object->prev_glob = ActiveObjects;
        ActiveObjects->next_glob = object;
     }

/***************************************************************************\
* Function Name: UN_DeActivateObject
* Description  : Removes the object from the ActiveObjects list
* Arguments    : the object to deactivate
* Returns      : void
* Notes        : Assumes ActiveObjects list has a dummy object on top.
\***************************************************************************/
void UN_DeActivateObject(UN_Object* object)
     {
        /* connect previous object to next object */
        object->prev_glob->next_glob = object->next_glob;
        if (object->next_glob != NULL) /* if we are not last object in list */
          {
             /* connect next object to previous object */
             object->next_glob->prev_glob = object->prev_glob;
          }
     }


/***************************************************************************\
* Function Name: UN_AntiTanVel
* Description  : Takes the x and y components of a velocity and returns
*                the direction number that an object with that velocity
*                would be heading.
* Arguments    : x, y - the x,y velocities
* Returns      : void
* Notes        : Assumes ActiveObjects list has a dummy object on top.
\***************************************************************************/
UN_Direction UN_AntiTanVel (UN_Velocity x, UN_Velocity y)
{
        UN_Direction start, final, current;
   UN_Direction best;
        float ratio, maxdif;

        if (y==0) { /* needed since ratio assignment, and thus the for
                        loop below, won't work if y==0 */
                if (x>=0) return 0;
                else return NUM_DIRS/2;
        } else ratio = x/y;

/* direction 0 = up (positive y), NUM_DIRS/4 = right (positive x),
** NUMDIRS/2 = down (negative y), 3*(NUM_DIRS/4) = left (negative x) */

/* Use a quick sign check on x and y to remove 3/4ths of the number of
** directions that the for loop below has to check. */

        if (x>=0) {
                if (y>0) { /* +x, +y, 0 <= dir <= NUM_DIRS/4 */
                        start = 0; final = NUM_DIRS/4;
                } else { /* +x, -y, NUM_DIRS/4 <= dir <= NUM_DIRS/2 */
                        start = NUM_DIRS/4; final = NUM_DIRS/2;
                }
        } else {
                if (y>0) { /* -x, +y, 3*(NUM_DIRS/4) <= dir <= NUM_DIRS */
                        start = 3*(NUM_DIRS/4); final = NUM_DIRS;
                } else { /* -x, -y, NUM_DIRS/2 <= dir <= 3*(NUM_DIRS/4) */
                        start = NUM_DIRS/2; final = 3*(NUM_DIRS/4);
                }
        }
        best = start; maxdif = ABS(ratio - (UN_Cos[start]/UN_Sin[start]));

/* Explanation of for loop: try to find the direction number for which
** cos/sin is closest to the ratio of x/y.  This will be the direction
** number (0=up, etc. as explained above) closest to the heading
** desired. */

        for (current=start+1 ; current<=final ; current++)
                if ( maxdif > ABS(ratio - (UN_Cos[current] / UN_Sin[current])) )
                {  maxdif = ABS(ratio - (UN_Cos[current] / UN_Sin[current]));
                        best = current; }
        return best;
}

/***************************************************************************\
* Function Name: UN_NewPlayer
* Description  : Assigns global variable Player to a new player object when
*                the current player object is about to die.  If there are no
*                other player objects, does nothing.
* Arguments    : none
* Returns      : void
* Notes        : Currently assumes that there will never be more than one
*                player object, and thus always does nothing.
\***************************************************************************/
void UN_NewPlayer() {
        return;
}

/***************************************************************************\
* Function Name: UN_KillObject
* Description  : Handles the death of an object:
*                Players, probes, supplys, and eggs become food
*                Everything else is deleted
* Arguments    : The dead object
* Returns      : void
\***************************************************************************/
void UN_KillObject(UN_Object *object) {
        if (object->what == OBJ_PLAYER) {
                object->shield = object->mass;
                object->what = OBJ_D_PLAYER;
        } else if (object->what == OBJ_PROBE) {
                object->shield = object->mass;
                object->what = OBJ_D_PROBE;
        } else if (object->what == OBJ_SUPPLY) {
                object->shield = object->mass;
                object->what = OBJ_D_SUPPLY;
        } else if ((object->what == OBJ_UF_EGG)||(object->what == OBJ_F_EGG)) {
                object->shield = object->mass;
                object->what = OBJ_D_EGG;
        } else UN_DeleteObject(object);
        AS_PlayFX(FX_DIE);
}
